import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.proto import profiles_pb2
from graphsignal.inference_span import InferenceSpan
from graphsignal.tracers.tensorflow import TensorflowProfiler
from graphsignal.usage.process_reader import ProcessReader
from graphsignal.usage.nvml_reader import NvmlReader
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class InferenceSpanTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            workload_name='r1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    @patch.object(TensorflowProfiler, 'start', return_value=True)
    @patch.object(TensorflowProfiler, 'stop', return_value=True)
    @patch.object(ProcessReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_profile')
    @patch('time.time', return_value=1000)
    def test_start_stop(self, mocked_time, mocked_upload_profile, mocked_nvml_read, mocked_host_read,
                        mocked_stop, mocked_start):
        for i in range(10):
            span = InferenceSpan(
                model_name='m1',
                metadata={'k1': 'v2', 'k3': 3.0},
                operation_profiler=TensorflowProfiler())
            span.set_count('items', 256)
            span.add_metadata('k4', 'v4')
            time.sleep(0.01)
            span.stop()

        self.assertEqual(mocked_start.call_count, 2)
        self.assertEqual(mocked_stop.call_count, 2)
        profile = mocked_upload_profile.call_args[0][0]

        self.assertEqual(profile.model_name, 'm1')
        self.assertTrue(profile.worker_id != '')
        self.assertEqual(profile.workload_id, '5573e39b6600')
        self.assertTrue(profile.start_us > 0)
        self.assertTrue(profile.end_us > 0)
        self.assertEqual(len(profile.inference_stats.inference_counter.buckets_sec), 1)
        self.assertEqual(len(profile.inference_stats.extra_counters['items'].buckets_sec), 1)
        self.assertEqual(len(profile.inference_stats.time_reservoir_us), 8)
        self.assertEqual(profile.metadata[0].key, 'k1')
        self.assertEqual(profile.metadata[0].value, 'v2')
        self.assertEqual(profile.metadata[1].key, 'k3')
        self.assertEqual(profile.metadata[1].value, '3.0')
        self.assertEqual(profile.metadata[2].key, 'k4')
        self.assertEqual(profile.metadata[2].value, 'v4')

    @patch.object(TensorflowProfiler, 'start', return_value=True)
    @patch.object(TensorflowProfiler, 'stop', return_value=True)
    @patch.object(ProcessReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_profile')
    def test_start_exception(self, mocked_upload_profile, mocked_nvml_read, mocked_host_read,
                             mocked_stop, mocked_start):
        mocked_start.side_effect = Exception('ex1')
        span = InferenceSpan(
            model_name='m1',
            ensure_profile=True,
            operation_profiler=TensorflowProfiler())
        span.stop()

        mocked_start.assert_called_once()
        mocked_stop.assert_not_called()
        profile = mocked_upload_profile.call_args[0][0]

        self.assertEqual(profile.model_name, 'm1')
        self.assertTrue(profile.worker_id != '')
        self.assertTrue(profile.start_us > 0)
        self.assertTrue(profile.end_us > 0)
        self.assertEqual(profile.profiler_errors[0].message, 'ex1')
        self.assertNotEqual(profile.profiler_errors[0].stack_trace, '')

    @patch.object(TensorflowProfiler, 'start', return_value=True)
    @patch.object(TensorflowProfiler, 'stop', return_value=True)
    @patch.object(ProcessReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_profile')
    def test_profiler_exception(self, mocked_upload_profile, mocked_nvml_read, mocked_host_read,
                            mocked_stop, mocked_start):
        mocked_stop.side_effect = Exception('ex1')
        span = InferenceSpan(
            model_name='m1',
            ensure_profile=True,
            operation_profiler=TensorflowProfiler())
        span.stop()

        mocked_start.assert_called_once()
        mocked_stop.assert_called_once()
        profile = mocked_upload_profile.call_args[0][0]

        self.assertEqual(profile.model_name, 'm1')
        self.assertTrue(profile.worker_id != '')
        self.assertTrue(profile.start_us > 0)
        self.assertTrue(profile.end_us > 0)
        self.assertEqual(profile.profiler_errors[0].message, 'ex1')
        self.assertNotEqual(profile.profiler_errors[0].stack_trace, '')

    @patch.object(TensorflowProfiler, 'start', return_value=True)
    @patch.object(TensorflowProfiler, 'stop', return_value=True)
    @patch.object(ProcessReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_profile')
    def test_inference_exception(self, mocked_upload_profile, mocked_nvml_read, mocked_host_read,
                            mocked_stop, mocked_start):

        for _ in range(2):
            try:
                with InferenceSpan(model_name='m1', ensure_profile=True, operation_profiler=TensorflowProfiler()):
                    raise Exception('ex1')
            except Exception as ex:
                if str(ex) != 'ex1':
                    raise ex

        profile = mocked_upload_profile.call_args[0][0]

        self.assertEqual(profile.model_name, 'm1')
        self.assertTrue(profile.worker_id != '')
        self.assertTrue(profile.start_us > 0)
        self.assertTrue(profile.end_us > 0)
        self.assertEqual(len(profile.inference_stats.exception_counter.buckets_sec), 1)
        self.assertEqual(profile.exception_stats[0].exc_type, 'Exception')
        self.assertEqual(profile.exception_stats[0].message, 'ex1')
        self.assertNotEqual(profile.exception_stats[0].stack_trace, '')
        self.assertEqual(profile.exception_stats[0].count, 2)

    @patch.object(TensorflowProfiler, 'start', return_value=True)
    @patch.object(TensorflowProfiler, 'stop', return_value=True)
    @patch.object(ProcessReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_profile')
    def test_add_exception(self, mocked_upload_profile, mocked_nvml_read, mocked_host_read,
                            mocked_stop, mocked_start):

        span = InferenceSpan(model_name='m1', ensure_profile=True, operation_profiler=TensorflowProfiler())
        try:
            raise Exception('ex2')
        except Exception as ex:
            span.add_exception(exc_info=True)
        span.stop()

        profile = mocked_upload_profile.call_args[0][0]

        self.assertEqual(profile.model_name, 'm1')
        self.assertTrue(profile.worker_id != '')
        self.assertTrue(profile.start_us > 0)
        self.assertTrue(profile.end_us > 0)
        self.assertEqual(len(profile.inference_stats.exception_counter.buckets_sec), 1)
        self.assertEqual(profile.exception_stats[0].exc_type, 'Exception')
        self.assertEqual(profile.exception_stats[0].message, 'ex2')
        self.assertNotEqual(profile.exception_stats[0].stack_trace, '')
        self.assertEqual(profile.exception_stats[0].count, 1)



from typing import Optional
import logging
import time
from torch import Tensor
import pytorch_lightning
from pytorch_lightning.callbacks.base import Callback

import graphsignal
from graphsignal.proto import profiles_pb2
from graphsignal.proto_utils import parse_semver
from graphsignal.tracers.pytorch import PyTorchProfiler
from graphsignal.inference_span import InferenceSpan

logger = logging.getLogger('graphsignal')


class GraphsignalCallback(Callback):
    def __init__(self, model_name: str, metadata: Optional[dict] = None, batch_size: Optional[int] = None):
        super().__init__()
        self._pl_version = None
        self._profiler = PyTorchProfiler()
        self._span = None
        self._model_name = model_name
        self._metadata = metadata
        self._batch_size = batch_size
        self._model_size_mb = None
        self._node_rank = -1
        self._local_rank = -1
        self._global_rank = -1

    def on_validate_start(self, trainer, pl_module):
        self._configure_profiler(trainer, pl_module)

    def on_validate_end(self, trainer, pl_module):
        graphsignal.upload()

    def on_validate_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        self._start_profiler(trainer)

    def on_validate_batch_end(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        self._stop_profiler(trainer)

    def on_test_start(self, trainer, pl_module):
        self._configure_profiler(trainer, pl_module)

    def on_test_end(self, trainer, pl_module):
        graphsignal.upload()

    def on_test_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        self._start_profiler(trainer)

    def on_test_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        self._stop_profiler(trainer)

    def on_predict_start(self, trainer, pl_module):
        self._configure_profiler(trainer, pl_module)

    def on_predict_end(self, trainer, pl_module):
        graphsignal.upload()

    def on_predict_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        self._start_profiler(trainer)

    def on_predict_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        self._stop_profiler(trainer)

    def _configure_profiler(self, trainer, pl_module):
        try:
            self._pl_version = profiles_pb2.SemVer()
            parse_semver(self._pl_version, pytorch_lightning.__version__)

            if self._check_param(trainer, 'node_rank') and trainer.node_rank >= 0:
                self._node_rank = trainer.node_rank
            if self._check_param(trainer, 'local_rank') and trainer.local_rank >= 0:
                self._local_rank = trainer.local_rank
            if self._check_param(trainer, 'global_rank') and trainer.global_rank >= 0:
                self._global_rank = trainer.global_rank

            model_size_mb = pytorch_lightning.utilities.memory.get_model_size_mb(pl_module)
            if model_size_mb:
                self._model_size_mb = model_size_mb
        except Exception:
            logger.error('Error configuring PyTorch Lightning profiler', exc_info=True)

    def _start_profiler(self, trainer):
        if not self._span:
            self._span = InferenceSpan(
                model_name=self._model_name,
                metadata=self._metadata,
                operation_profiler=self._profiler)
            self._span.set_count('items', self._batch_size)

    def _stop_profiler(self, trainer):
        if self._span:
            if self._span._is_scheduled:
                self._update_profile(trainer)
            self._span.stop()
            self._span = None

    def _update_profile(self, trainer):
        try:
            profile = self._span._profile

            profile.profiler_info.framework_profiler_type = profiles_pb2.ProfilerInfo.ProfilerType.PYTORCH_LIGHTNING_PROFILER

            framework = profile.frameworks.add()
            framework.type = profiles_pb2.FrameworkInfo.FrameworkType.PYTORCH_LIGHTNING_FRAMEWORK
            framework.version.CopyFrom(self._pl_version)

            if self._check_param(trainer, 'world_size'):
                profile.cluster_info.world_size = trainer.world_size

            profile.model_info.model_format = profiles_pb2.ModelInfo.ModelFormat.PYTORCH_FORMAT
            if self._model_size_mb:
                profile.model_info.model_size_bytes = int(self._model_size_mb * 1e6)

            if self._node_rank >= 0:
                profile.node_usage.node_rank = self._node_rank
            if self._global_rank >= 0:
                profile.process_usage.global_rank = self._global_rank
            if self._local_rank >= 0:
                profile.process_usage.local_rank = self._local_rank
        except Exception as exc:
            self._span._add_profiler_exception(exc)

    def _check_param(self, trainer, param):
        value = getattr(trainer, param, None)
        return isinstance(value, (str, int, float, bool))




class Agent:
    def __init__(self):
        self.worker_id = None
        self.api_key = None
        self.workload_name = None
        self.debug_mode = None
        self.disable_op_profiler = False
        self.uploader = None
        self.process_reader = None
        self.nvml_reader = None
        self.workload = None



from typing import Union, Any
import logging
import sys
import time
from threading import Lock
import traceback

import graphsignal
from graphsignal.proto import profiles_pb2

logger = logging.getLogger('graphsignal')


class InferenceSpan:

    __slots__ = [
        '_operation_profiler',
        '_profile_scheduler',
        '_inference_stats',
        '_workload',
        '_model_name',
        '_ensure_profile',
        '_metadata',
        '_context',
        '_is_scheduled',
        '_is_profiling',
        '_profile',
        '_is_stopped',
        '_start_counter',
        '_extra_counts'
    ]

    def __init__(self, 
            model_name,
            metadata=None,
            ensure_profile=False, 
            operation_profiler=None,
            context=None):
        if not model_name:
            raise ValueError('model_name is required')
        if metadata is not None and not isinstance(metadata, dict):
            raise ValueError('metadata must be dict')
        self._model_name = model_name
        self._metadata = metadata
        self._ensure_profile = ensure_profile
        self._operation_profiler = operation_profiler
        self._context = context

        self._workload = None
        self._profile_scheduler = None
        self._inference_stats = None
        self._is_stopped = False
        self._is_scheduled = False
        self._is_profiling = False
        self._profile = None
        self._extra_counts = None

        try:
            self._start()
        except Exception as ex:
            if self._is_scheduled:
                self._is_stopped = True
                self._profile_scheduler.unlock()
            raise ex

    def _start(self):
        if self._is_stopped:
            return

        self._workload = graphsignal.workload()
        self._profile_scheduler = self._workload.get_profile_scheduler(self._model_name)
        self._inference_stats = self._workload.get_inference_stats(self._model_name)

        if self._profile_scheduler.lock(ensure=self._ensure_profile):
            if logger.isEnabledFor(logging.DEBUG):
                profiling_start_overhead_counter = time.perf_counter()

            self._is_scheduled = True
            self._profile = self._workload.create_profile()
            self._profile.model_name = self._model_name

            if not graphsignal._agent.disable_op_profiler and self._operation_profiler:
                try:
                    self._operation_profiler.start(self._profile, self._context)
                    self._is_profiling = True
                except Exception as exc:
                    self._add_profiler_exception(exc)
                    logger.error('Error starting profiler', exc_info=True)

            self._profile.start_us = _timestamp_us()

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('Profiling start took: %fs', time.perf_counter() - profiling_start_overhead_counter)

        self._start_counter = time.perf_counter()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.stop(exc_info=exc_info)
        return False

    def stop(self, exc_info=None) -> None:
        try:
            self._stop(exc_info=exc_info)
        finally:
            if self._is_scheduled:
                self._is_stopped = True
                self._profile_scheduler.unlock()            

    def _stop(self, exc_info=None) -> None:
        stop_counter = time.perf_counter()
        end_us = _timestamp_us()

        if self._is_stopped:
            return
        self._is_stopped = True

        if self._is_scheduled:
            self._profile.end_us = end_us

            if self._is_profiling:
                try:
                    self._operation_profiler.stop(self._profile, self._context)
                except Exception as exc:
                    logger.error('Error stopping profiler', exc_info=True)
                    self._add_profiler_exception(exc)

        self._inference_stats.inc_inference_counter(1, end_us)
        if self._extra_counts is not None:
            for name, value in self._extra_counts.items():
                self._inference_stats.inc_extra_counter(name, value, end_us)

        # only measure time if not profiling to exclude profiler overhead
        if not self._is_profiling:
            self._inference_stats.add_time(int((stop_counter - self._start_counter) * 1e6))

        if exc_info and exc_info[0]:
            self._inference_stats.inc_exception_counter(1, end_us)
            self._workload.add_exception(exc_info, end_us)

        if self._is_scheduled:
            if logger.isEnabledFor(logging.DEBUG):
                profiling_stop_overhead_counter = time.perf_counter()

            try:
                graphsignal._agent.process_reader.read(self._profile)
                graphsignal._agent.nvml_reader.read(self._profile)
            except Exception as exc:
                logger.error('Error reading usage information', exc_info=True)
                self._add_profiler_exception(exc)

            self._inference_stats.finalize(end_us)
            for time_us in self._inference_stats.time_reservoir_us:
                self._profile.inference_stats.time_reservoir_us.append(time_us)
            inference_counter_proto = self._profile.inference_stats.inference_counter
            for bucket, count in self._inference_stats.inference_counter.items():
                inference_counter_proto.buckets_sec[bucket] = count
            exception_counter_proto = self._profile.inference_stats.exception_counter
            for bucket, count in self._inference_stats.exception_counter.items():
                exception_counter_proto.buckets_sec[bucket] = count
            for name, counter in self._inference_stats.extra_counters.items():
                extra_counter_proto = self._profile.inference_stats.extra_counters[name]
                for bucket, count in counter.items():
                    extra_counter_proto.buckets_sec[bucket] = count
            self._workload.reset_inference_stats(self._model_name)

            if self._metadata is not None:
                for key, value in self._metadata.items():
                    entry = self._profile.metadata.add()
                    entry.key = key
                    entry.value = str(value)

            for exc_stats in self._workload.exception_stats.values():
                self._profile.exception_stats.append(exc_stats)

            graphsignal._agent.uploader.upload_profile(self._profile)
            self._workload.tick()

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('Profiling stop took: %fs', time.perf_counter() - profiling_stop_overhead_counter)

    def set_count(self, name: str, value: Union[int, float]) -> None:
        if not name:
            raise ValueError('set_count: name must be provided')
        if not isinstance(value, (int, float)):
            raise ValueError('set_count: value must be int or float')

        if self._extra_counts is None:
            self._extra_counts = {}
        self._extra_counts[name] = value

    def add_metadata(self, key: str, value: Any) -> None:
        if not key:
            raise ValueError('add_metadata: key must be provided')
        if value is None:
            raise ValueError('add_metadata: value must be provided')

        if self._metadata is None:
            self._metadata = {}
        self._metadata[key] = value

    def add_exception(self, exc_info=Union[bool, tuple]) -> None:
        if not exc_info:
            raise ValueError('add_exception: exc_info must be provided')

        if exc_info == True:
            exc_info = sys.exc_info()

        self._workload.add_exception(exc_info, _timestamp_us())

    def _add_profiler_exception(self, exc):
        profiler_error = self._profile.profiler_errors.add()
        profiler_error.message = str(exc)
        if exc.__traceback__:
            frames = traceback.format_tb(exc.__traceback__)
            if len(frames) > 0:
                profiler_error.stack_trace = ''.join(frames)


def _timestamp_us():
    return int(time.time() * 1e6)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: profiles.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0eprofiles.proto\x12\x14graphsignal.profiles\"X\n\rUploadRequest\x12\x34\n\x0bml_profiles\x18\x01 \x03(\x0b\x32\x1f.graphsignal.profiles.MLProfile\x12\x11\n\tupload_ms\x18\x02 \x01(\x04\"\x10\n\x0eUploadResponse\"\xad\x01\n\rProfileRecord\x12\x0f\n\x07\x64\x61ta_id\x18\x01 \x01(\t\x12\x10\n\x08model_id\x18\x02 \x01(\t\x12\x12\n\nprofile_id\x18\x03 \x01(\t\x12\x33\n\nml_profile\x18\x04 \x01(\x0b\x32\x1f.graphsignal.profiles.MLProfile\x12\x1a\n\x12\x64\x61ta_retention_sec\x18\x05 \x01(\x04\x12\x14\n\x0ctime_skew_ms\x18\x06 \x01(\x12\"\xd1\x07\n\tMLProfile\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x15\n\rworkload_name\x18) \x01(\t\x12\x11\n\tworker_id\x18\x1c \x01(\t\x12\x13\n\x0bworkload_id\x18\x02 \x01(\t\x12\x10\n\x08start_us\x18\x05 \x01(\x04\x12\x0e\n\x06\x65nd_us\x18\x06 \x01(\x04\x12\x30\n\x08metadata\x18- \x03(\x0b\x32\x1e.graphsignal.profiles.Metadata\x12\x37\n\nframeworks\x18$ \x03(\x0b\x32#.graphsignal.profiles.FrameworkInfo\x12/\n\x08op_stats\x18\x0b \x03(\x0b\x32\x1d.graphsignal.profiles.OpStats\x12\x37\n\x0ckernel_stats\x18\x0c \x03(\x0b\x32!.graphsignal.profiles.KernelStats\x12\x12\n\ntrace_data\x18! \x01(\x0c\x12\x1a\n\x12is_trace_available\x18\" \x01(\x08\x12=\n\x0finference_stats\x18* \x01(\x0b\x32$.graphsignal.profiles.InferenceStats\x12=\n\x0f\x65xception_stats\x18. \x03(\x0b\x32$.graphsignal.profiles.ExceptionStats\x12\x33\n\nmodel_info\x18& \x01(\x0b\x32\x1f.graphsignal.profiles.ModelInfo\x12\x37\n\x0c\x63luster_info\x18+ \x01(\x0b\x32!.graphsignal.profiles.ClusterInfo\x12<\n\ncomm_usage\x18\x1b \x01(\x0b\x32(.graphsignal.profiles.CommunicationUsage\x12\x37\n\x0c\x64\x65vice_usage\x18\x12 \x03(\x0b\x32!.graphsignal.profiles.DeviceUsage\x12\x33\n\nnode_usage\x18\x19 \x01(\x0b\x32\x1f.graphsignal.profiles.NodeUsage\x12\x39\n\rprocess_usage\x18\x1a \x01(\x0b\x32\".graphsignal.profiles.ProcessUsage\x12\x39\n\rprofiler_info\x18# \x01(\x0b\x32\".graphsignal.profiles.ProfilerInfo\x12<\n\x0fprofiler_errors\x18\x07 \x03(\x0b\x32#.graphsignal.profiles.ProfilerError\"5\n\x06SemVer\x12\r\n\x05major\x18\x01 \x01(\x05\x12\r\n\x05minor\x18\x02 \x01(\x05\x12\r\n\x05patch\x18\x03 \x01(\x05\"&\n\x08Metadata\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\xb8\x02\n\rFrameworkInfo\x12?\n\x04type\x18\x01 \x01(\x0e\x32\x31.graphsignal.profiles.FrameworkInfo.FrameworkType\x12-\n\x07version\x18\x02 \x01(\x0b\x32\x1c.graphsignal.profiles.SemVer\"\xb6\x01\n\rFrameworkType\x12\x17\n\x13\x46RAMEWORK_UNDEFINED\x10\x00\x12\x18\n\x14TENSORFLOW_FRAMEWORK\x10\x01\x12\x15\n\x11PYTORCH_FRAMEWORK\x10\x02\x12\x13\n\x0fKERAS_FRAMEWORK\x10\x03\x12\x1f\n\x1bPYTORCH_LIGHTNING_FRAMEWORK\x10\x04\x12\x11\n\rJAX_FRAMEWORK\x10\x07\x12\x12\n\x0eONNX_FRAMEWORK\x10\t\"\xea\x03\n\x07OpStats\x12\x35\n\x0b\x64\x65vice_type\x18\x01 \x01(\x0e\x32 .graphsignal.profiles.DeviceType\x12\x11\n\tdevice_id\x18\x02 \x01(\t\x12\x0f\n\x07op_type\x18\x03 \x01(\t\x12\x0f\n\x07op_name\x18\x04 \x01(\t\x12\r\n\x05\x63ount\x18\x05 \x01(\x04\x12\x1a\n\x12total_host_time_us\x18\x06 \x01(\x04\x12\x1c\n\x14total_device_time_us\x18\x07 \x01(\x04\x12\x19\n\x11self_host_time_us\x18\x08 \x01(\x04\x12\x1b\n\x13self_device_time_us\x18\t \x01(\x04\x12\x19\n\x11total_host_memory\x18\n \x01(\x04\x12\x1b\n\x13total_device_memory\x18\x0b \x01(\x04\x12\x18\n\x10self_host_memory\x18\x0c \x01(\x04\x12\x1a\n\x12self_device_memory\x18\r \x01(\x04\x12\x1d\n\x15self_host_memory_rate\x18\x0e \x01(\x04\x12\x1f\n\x17self_device_memory_rate\x18\x0f \x01(\x04\x12\r\n\x05\x66lops\x18\x10 \x01(\x04\x12\x15\n\rflops_per_sec\x18\x12 \x01(\x01\x12\x1e\n\x16tensorcore_utilization\x18\x11 \x01(\x01\"\xbe\x01\n\x0bKernelStats\x12\x35\n\x0b\x64\x65vice_type\x18\x01 \x01(\x0e\x32 .graphsignal.profiles.DeviceType\x12\x11\n\tdevice_id\x18\x02 \x01(\t\x12\x0f\n\x07op_name\x18\x03 \x01(\t\x12\x13\n\x0bkernel_name\x18\x04 \x01(\t\x12\r\n\x05\x63ount\x18\x05 \x01(\x04\x12\x13\n\x0b\x64uration_ns\x18\x06 \x01(\x04\x12\x1b\n\x13is_using_tensorcore\x18\x07 \x01(\x08\"\x84\x04\n\x0eInferenceStats\x12\x19\n\x11time_reservoir_us\x18\x01 \x03(\x04\x12G\n\x11inference_counter\x18\x02 \x01(\x0b\x32,.graphsignal.profiles.InferenceStats.Counter\x12G\n\x11\x65xception_counter\x18\x03 \x01(\x0b\x32,.graphsignal.profiles.InferenceStats.Counter\x12O\n\x0e\x65xtra_counters\x18\x04 \x03(\x0b\x32\x37.graphsignal.profiles.InferenceStats.ExtraCountersEntry\x1a\x8f\x01\n\x07\x43ounter\x12Q\n\x0b\x62uckets_sec\x18\x01 \x03(\x0b\x32<.graphsignal.profiles.InferenceStats.Counter.BucketsSecEntry\x1a\x31\n\x0f\x42ucketsSecEntry\x12\x0b\n\x03key\x18\x01 \x01(\x04\x12\r\n\x05value\x18\x02 \x01(\x04:\x02\x38\x01\x1a\x62\n\x12\x45xtraCountersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12;\n\x05value\x18\x02 \x01(\x0b\x32,.graphsignal.profiles.InferenceStats.Counter:\x02\x38\x01\"j\n\x0e\x45xceptionStats\x12\x10\n\x08\x65xc_type\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x13\n\x0bstack_trace\x18\x03 \x01(\t\x12\r\n\x05\x63ount\x18\x04 \x01(\x04\x12\x11\n\tupdate_ms\x18\x05 \x01(\x04\"\xb2\x01\n\tModelInfo\x12\x41\n\x0cmodel_format\x18\x01 \x01(\x0e\x32+.graphsignal.profiles.ModelInfo.ModelFormat\x12\x18\n\x10model_size_bytes\x18\x02 \x01(\x04\"H\n\x0bModelFormat\x12\x14\n\x10\x46ORMAT_UNDEFINED\x10\x00\x12\x12\n\x0ePYTORCH_FORMAT\x10\x01\x12\x0f\n\x0bONNX_FORMAT\x10\x02\"!\n\x0b\x43lusterInfo\x12\x12\n\nworld_size\x18\x01 \x01(\x04\"\xd5\x01\n\x12\x43ommunicationUsage\x12W\n\x0c\x62\x61\x63kend_type\x18\x01 \x01(\x0e\x32\x41.graphsignal.profiles.CommunicationUsage.CommunicationBackendType\"f\n\x18\x43ommunicationBackendType\x12#\n\x1f\x43OMMUNICATION_BACKEND_UNDEFINED\x10\x00\x12\x08\n\x04NCCL\x10\x01\x12\x08\n\x04GLOO\x10\x02\x12\x07\n\x03MPI\x10\x03\x12\x08\n\x04RING\x10\x04\"\xc6\x01\n\tNodeUsage\x12\x10\n\x08hostname\x18\x01 \x01(\t\x12\x12\n\nip_address\x18\x07 \x01(\t\x12\x11\n\tnode_rank\x18\x08 \x01(\x05\x12\x10\n\x08mem_used\x18\t \x01(\x04\x12\x11\n\tmem_total\x18\n \x01(\x04\x12\x10\n\x08platform\x18\x02 \x01(\t\x12\x0f\n\x07machine\x18\x03 \x01(\t\x12\x0f\n\x07os_name\x18\x04 \x01(\t\x12\x12\n\nos_version\x18\x05 \x01(\t\x12\x13\n\x0bnum_devices\x18\x06 \x01(\x05\"\xf9\x02\n\x0cProcessUsage\x12\x12\n\nprocess_id\x18\x05 \x01(\t\x12\x10\n\x08start_ms\x18\x0e \x01(\x04\x12\x12\n\nlocal_rank\x18\x0c \x01(\x05\x12\x13\n\x0bglobal_rank\x18\x0f \x01(\x05\x12\x10\n\x08\x63pu_name\x18\x10 \x01(\t\x12\x19\n\x11\x63pu_usage_percent\x18\x01 \x01(\x01\x12\x0f\n\x07max_rss\x18\x02 \x01(\x04\x12\x13\n\x0b\x63urrent_rss\x18\x03 \x01(\x04\x12\x0f\n\x07vm_size\x18\x04 \x01(\x04\x12;\n\x07runtime\x18\x06 \x01(\x0e\x32*.graphsignal.profiles.ProcessUsage.Runtime\x12\x35\n\x0fruntime_version\x18\x07 \x01(\x0b\x32\x1c.graphsignal.profiles.SemVer\x12\x14\n\x0cruntime_impl\x18\x08 \x01(\t\",\n\x07Runtime\x12\x15\n\x11RUNTIME_UNDEFINED\x10\x00\x12\n\n\x06PYTHON\x10\x01\"\xcf\x03\n\x0b\x44\x65viceUsage\x12\x35\n\x0b\x64\x65vice_type\x18\x0b \x01(\x0e\x32 .graphsignal.profiles.DeviceType\x12\x11\n\tdevice_id\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65vice_name\x18\x02 \x01(\t\x12\x14\n\x0c\x61rchitecture\x18\x11 \x01(\t\x12\x38\n\x12\x63ompute_capability\x18\x0f \x01(\x0b\x32\x1c.graphsignal.profiles.SemVer\x12\x11\n\tmem_total\x18\x03 \x01(\x04\x12\x10\n\x08mem_used\x18\x04 \x01(\x04\x12\x10\n\x08mem_free\x18\x05 \x01(\x04\x12\x1f\n\x17gpu_utilization_percent\x18\x06 \x01(\x01\x12\x1a\n\x12mem_access_percent\x18\x10 \x01(\x01\x12\x1a\n\x12pcie_throughput_tx\x18\r \x01(\x01\x12\x1a\n\x12pcie_throughput_rx\x18\x0e \x01(\x01\x12\x12\n\ngpu_temp_c\x18\x08 \x01(\x01\x12\x15\n\rpower_usage_w\x18\t \x01(\x01\x12\x19\n\x11\x66\x61n_speed_percent\x18\n \x01(\x01\x12\x1f\n\x17mxu_utilization_percent\x18\x0c \x01(\x01\"\xa8\x03\n\x0cProfilerInfo\x12P\n\x17operation_profiler_type\x18\x01 \x01(\x0e\x32/.graphsignal.profiles.ProfilerInfo.ProfilerType\x12P\n\x17\x66ramework_profiler_type\x18\x02 \x01(\x0e\x32/.graphsignal.profiles.ProfilerInfo.ProfilerType\x12-\n\x07version\x18\x03 \x01(\x0b\x32\x1c.graphsignal.profiles.SemVer\"\xc4\x01\n\x0cProfilerType\x12\x16\n\x12PROFILER_UNDEFINED\x10\x00\x12\x14\n\x10GENERIC_PROFILER\x10\x01\x12\x17\n\x13TENSORFLOW_PROFILER\x10\x02\x12\x14\n\x10PYTORCH_PROFILER\x10\x03\x12\x12\n\x0eKERAS_PROFILER\x10\x04\x12\x1e\n\x1aPYTORCH_LIGHTNING_PROFILER\x10\x06\x12\x10\n\x0cJAX_PROFILER\x10\x08\x12\x11\n\rONNX_PROFILER\x10\n\"5\n\rProfilerError\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x13\n\x0bstack_trace\x18\x02 \x01(\t*B\n\nDeviceType\x12\x19\n\x15\x44\x45VICE_TYPE_UNDEFINED\x10\x00\x12\x07\n\x03\x43PU\x10\x01\x12\x07\n\x03GPU\x10\x02\x12\x07\n\x03TPU\x10\x03\x62\x06proto3')

_DEVICETYPE = DESCRIPTOR.enum_types_by_name['DeviceType']
DeviceType = enum_type_wrapper.EnumTypeWrapper(_DEVICETYPE)
DEVICE_TYPE_UNDEFINED = 0
CPU = 1
GPU = 2
TPU = 3


_UPLOADREQUEST = DESCRIPTOR.message_types_by_name['UploadRequest']
_UPLOADRESPONSE = DESCRIPTOR.message_types_by_name['UploadResponse']
_PROFILERECORD = DESCRIPTOR.message_types_by_name['ProfileRecord']
_MLPROFILE = DESCRIPTOR.message_types_by_name['MLProfile']
_SEMVER = DESCRIPTOR.message_types_by_name['SemVer']
_METADATA = DESCRIPTOR.message_types_by_name['Metadata']
_FRAMEWORKINFO = DESCRIPTOR.message_types_by_name['FrameworkInfo']
_OPSTATS = DESCRIPTOR.message_types_by_name['OpStats']
_KERNELSTATS = DESCRIPTOR.message_types_by_name['KernelStats']
_INFERENCESTATS = DESCRIPTOR.message_types_by_name['InferenceStats']
_INFERENCESTATS_COUNTER = _INFERENCESTATS.nested_types_by_name['Counter']
_INFERENCESTATS_COUNTER_BUCKETSSECENTRY = _INFERENCESTATS_COUNTER.nested_types_by_name['BucketsSecEntry']
_INFERENCESTATS_EXTRACOUNTERSENTRY = _INFERENCESTATS.nested_types_by_name['ExtraCountersEntry']
_EXCEPTIONSTATS = DESCRIPTOR.message_types_by_name['ExceptionStats']
_MODELINFO = DESCRIPTOR.message_types_by_name['ModelInfo']
_CLUSTERINFO = DESCRIPTOR.message_types_by_name['ClusterInfo']
_COMMUNICATIONUSAGE = DESCRIPTOR.message_types_by_name['CommunicationUsage']
_NODEUSAGE = DESCRIPTOR.message_types_by_name['NodeUsage']
_PROCESSUSAGE = DESCRIPTOR.message_types_by_name['ProcessUsage']
_DEVICEUSAGE = DESCRIPTOR.message_types_by_name['DeviceUsage']
_PROFILERINFO = DESCRIPTOR.message_types_by_name['ProfilerInfo']
_PROFILERERROR = DESCRIPTOR.message_types_by_name['ProfilerError']
_FRAMEWORKINFO_FRAMEWORKTYPE = _FRAMEWORKINFO.enum_types_by_name['FrameworkType']
_MODELINFO_MODELFORMAT = _MODELINFO.enum_types_by_name['ModelFormat']
_COMMUNICATIONUSAGE_COMMUNICATIONBACKENDTYPE = _COMMUNICATIONUSAGE.enum_types_by_name['CommunicationBackendType']
_PROCESSUSAGE_RUNTIME = _PROCESSUSAGE.enum_types_by_name['Runtime']
_PROFILERINFO_PROFILERTYPE = _PROFILERINFO.enum_types_by_name['ProfilerType']
UploadRequest = _reflection.GeneratedProtocolMessageType('UploadRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADREQUEST,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.UploadRequest)
  })
_sym_db.RegisterMessage(UploadRequest)

UploadResponse = _reflection.GeneratedProtocolMessageType('UploadResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADRESPONSE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.UploadResponse)
  })
_sym_db.RegisterMessage(UploadResponse)

ProfileRecord = _reflection.GeneratedProtocolMessageType('ProfileRecord', (_message.Message,), {
  'DESCRIPTOR' : _PROFILERECORD,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.ProfileRecord)
  })
_sym_db.RegisterMessage(ProfileRecord)

MLProfile = _reflection.GeneratedProtocolMessageType('MLProfile', (_message.Message,), {
  'DESCRIPTOR' : _MLPROFILE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.MLProfile)
  })
_sym_db.RegisterMessage(MLProfile)

SemVer = _reflection.GeneratedProtocolMessageType('SemVer', (_message.Message,), {
  'DESCRIPTOR' : _SEMVER,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.SemVer)
  })
_sym_db.RegisterMessage(SemVer)

Metadata = _reflection.GeneratedProtocolMessageType('Metadata', (_message.Message,), {
  'DESCRIPTOR' : _METADATA,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.Metadata)
  })
_sym_db.RegisterMessage(Metadata)

FrameworkInfo = _reflection.GeneratedProtocolMessageType('FrameworkInfo', (_message.Message,), {
  'DESCRIPTOR' : _FRAMEWORKINFO,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.FrameworkInfo)
  })
_sym_db.RegisterMessage(FrameworkInfo)

OpStats = _reflection.GeneratedProtocolMessageType('OpStats', (_message.Message,), {
  'DESCRIPTOR' : _OPSTATS,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.OpStats)
  })
_sym_db.RegisterMessage(OpStats)

KernelStats = _reflection.GeneratedProtocolMessageType('KernelStats', (_message.Message,), {
  'DESCRIPTOR' : _KERNELSTATS,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.KernelStats)
  })
_sym_db.RegisterMessage(KernelStats)

InferenceStats = _reflection.GeneratedProtocolMessageType('InferenceStats', (_message.Message,), {

  'Counter' : _reflection.GeneratedProtocolMessageType('Counter', (_message.Message,), {

    'BucketsSecEntry' : _reflection.GeneratedProtocolMessageType('BucketsSecEntry', (_message.Message,), {
      'DESCRIPTOR' : _INFERENCESTATS_COUNTER_BUCKETSSECENTRY,
      '__module__' : 'profiles_pb2'
      # @@protoc_insertion_point(class_scope:graphsignal.profiles.InferenceStats.Counter.BucketsSecEntry)
      })
    ,
    'DESCRIPTOR' : _INFERENCESTATS_COUNTER,
    '__module__' : 'profiles_pb2'
    # @@protoc_insertion_point(class_scope:graphsignal.profiles.InferenceStats.Counter)
    })
  ,

  'ExtraCountersEntry' : _reflection.GeneratedProtocolMessageType('ExtraCountersEntry', (_message.Message,), {
    'DESCRIPTOR' : _INFERENCESTATS_EXTRACOUNTERSENTRY,
    '__module__' : 'profiles_pb2'
    # @@protoc_insertion_point(class_scope:graphsignal.profiles.InferenceStats.ExtraCountersEntry)
    })
  ,
  'DESCRIPTOR' : _INFERENCESTATS,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.InferenceStats)
  })
_sym_db.RegisterMessage(InferenceStats)
_sym_db.RegisterMessage(InferenceStats.Counter)
_sym_db.RegisterMessage(InferenceStats.Counter.BucketsSecEntry)
_sym_db.RegisterMessage(InferenceStats.ExtraCountersEntry)

ExceptionStats = _reflection.GeneratedProtocolMessageType('ExceptionStats', (_message.Message,), {
  'DESCRIPTOR' : _EXCEPTIONSTATS,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.ExceptionStats)
  })
_sym_db.RegisterMessage(ExceptionStats)

ModelInfo = _reflection.GeneratedProtocolMessageType('ModelInfo', (_message.Message,), {
  'DESCRIPTOR' : _MODELINFO,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.ModelInfo)
  })
_sym_db.RegisterMessage(ModelInfo)

ClusterInfo = _reflection.GeneratedProtocolMessageType('ClusterInfo', (_message.Message,), {
  'DESCRIPTOR' : _CLUSTERINFO,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.ClusterInfo)
  })
_sym_db.RegisterMessage(ClusterInfo)

CommunicationUsage = _reflection.GeneratedProtocolMessageType('CommunicationUsage', (_message.Message,), {
  'DESCRIPTOR' : _COMMUNICATIONUSAGE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.CommunicationUsage)
  })
_sym_db.RegisterMessage(CommunicationUsage)

NodeUsage = _reflection.GeneratedProtocolMessageType('NodeUsage', (_message.Message,), {
  'DESCRIPTOR' : _NODEUSAGE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.NodeUsage)
  })
_sym_db.RegisterMessage(NodeUsage)

ProcessUsage = _reflection.GeneratedProtocolMessageType('ProcessUsage', (_message.Message,), {
  'DESCRIPTOR' : _PROCESSUSAGE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.ProcessUsage)
  })
_sym_db.RegisterMessage(ProcessUsage)

DeviceUsage = _reflection.GeneratedProtocolMessageType('DeviceUsage', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEUSAGE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.DeviceUsage)
  })
_sym_db.RegisterMessage(DeviceUsage)

ProfilerInfo = _reflection.GeneratedProtocolMessageType('ProfilerInfo', (_message.Message,), {
  'DESCRIPTOR' : _PROFILERINFO,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.ProfilerInfo)
  })
_sym_db.RegisterMessage(ProfilerInfo)

ProfilerError = _reflection.GeneratedProtocolMessageType('ProfilerError', (_message.Message,), {
  'DESCRIPTOR' : _PROFILERERROR,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.ProfilerError)
  })
_sym_db.RegisterMessage(ProfilerError)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _INFERENCESTATS_COUNTER_BUCKETSSECENTRY._options = None
  _INFERENCESTATS_COUNTER_BUCKETSSECENTRY._serialized_options = b'8\001'
  _INFERENCESTATS_EXTRACOUNTERSENTRY._options = None
  _INFERENCESTATS_EXTRACOUNTERSENTRY._serialized_options = b'8\001'
  _DEVICETYPE._serialized_start=4988
  _DEVICETYPE._serialized_end=5054
  _UPLOADREQUEST._serialized_start=40
  _UPLOADREQUEST._serialized_end=128
  _UPLOADRESPONSE._serialized_start=130
  _UPLOADRESPONSE._serialized_end=146
  _PROFILERECORD._serialized_start=149
  _PROFILERECORD._serialized_end=322
  _MLPROFILE._serialized_start=325
  _MLPROFILE._serialized_end=1302
  _SEMVER._serialized_start=1304
  _SEMVER._serialized_end=1357
  _METADATA._serialized_start=1359
  _METADATA._serialized_end=1397
  _FRAMEWORKINFO._serialized_start=1400
  _FRAMEWORKINFO._serialized_end=1712
  _FRAMEWORKINFO_FRAMEWORKTYPE._serialized_start=1530
  _FRAMEWORKINFO_FRAMEWORKTYPE._serialized_end=1712
  _OPSTATS._serialized_start=1715
  _OPSTATS._serialized_end=2205
  _KERNELSTATS._serialized_start=2208
  _KERNELSTATS._serialized_end=2398
  _INFERENCESTATS._serialized_start=2401
  _INFERENCESTATS._serialized_end=2917
  _INFERENCESTATS_COUNTER._serialized_start=2674
  _INFERENCESTATS_COUNTER._serialized_end=2817
  _INFERENCESTATS_COUNTER_BUCKETSSECENTRY._serialized_start=2768
  _INFERENCESTATS_COUNTER_BUCKETSSECENTRY._serialized_end=2817
  _INFERENCESTATS_EXTRACOUNTERSENTRY._serialized_start=2819
  _INFERENCESTATS_EXTRACOUNTERSENTRY._serialized_end=2917
  _EXCEPTIONSTATS._serialized_start=2919
  _EXCEPTIONSTATS._serialized_end=3025
  _MODELINFO._serialized_start=3028
  _MODELINFO._serialized_end=3206
  _MODELINFO_MODELFORMAT._serialized_start=3134
  _MODELINFO_MODELFORMAT._serialized_end=3206
  _CLUSTERINFO._serialized_start=3208
  _CLUSTERINFO._serialized_end=3241
  _COMMUNICATIONUSAGE._serialized_start=3244
  _COMMUNICATIONUSAGE._serialized_end=3457
  _COMMUNICATIONUSAGE_COMMUNICATIONBACKENDTYPE._serialized_start=3355
  _COMMUNICATIONUSAGE_COMMUNICATIONBACKENDTYPE._serialized_end=3457
  _NODEUSAGE._serialized_start=3460
  _NODEUSAGE._serialized_end=3658
  _PROCESSUSAGE._serialized_start=3661
  _PROCESSUSAGE._serialized_end=4038
  _PROCESSUSAGE_RUNTIME._serialized_start=3994
  _PROCESSUSAGE_RUNTIME._serialized_end=4038
  _DEVICEUSAGE._serialized_start=4041
  _DEVICEUSAGE._serialized_end=4504
  _PROFILERINFO._serialized_start=4507
  _PROFILERINFO._serialized_end=4931
  _PROFILERINFO_PROFILERTYPE._serialized_start=4735
  _PROFILERINFO_PROFILERTYPE._serialized_end=4931
  _PROFILERERROR._serialized_start=4933
  _PROFILERERROR._serialized_end=4986
# @@protoc_insertion_point(module_scope)



import unittest
import logging
import sys
import os
import time
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.workload import Workload

logger = logging.getLogger('graphsignal')


class WorkloadTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    @patch('time.time', return_value=1)
    def test_update_inference_stats(self, mocked_time):
        stats = Workload().get_inference_stats('m1')

        stats.inc_inference_counter(1, 1000 * 1e6)
        stats.inc_inference_counter(1, 1000 * 1e6)
        stats.inc_inference_counter(1, 1001 * 1e6)
        self.assertEqual(stats.inference_counter, {1: 0, 1000: 2, 1001: 1})
        
        stats.inc_exception_counter(1, 1003 * 1e6)
        stats.inc_exception_counter(1, 1003 * 1e6)
        stats.inc_exception_counter(1, 1004 * 1e6)
        self.assertEqual(stats.exception_counter, {1: 0, 1003: 2, 1004: 1})

        stats.inc_extra_counter('c1', 1, 1000 * 1e6)
        stats.inc_extra_counter('c1', 1, 1000 * 1e6)
        stats.inc_extra_counter('c1', 1, 1001 * 1e6)
        stats.inc_extra_counter('c2', 1, 1001 * 1e6)
        self.assertEqual(stats.extra_counters['c1'], {1: 0, 1000: 2, 1001: 1})
        self.assertEqual(stats.extra_counters['c2'], {1: 0, 1001: 1})

        stats.add_time(20)
        stats.add_time(30)
        self.assertEqual(stats.time_reservoir_us, [20, 30])




from typing import Optional
import logging
import os
import sys
import time
import gzip
import atexit
import threading
import onnxruntime

import graphsignal
from graphsignal.proto_utils import parse_semver
from graphsignal.proto import profiles_pb2
from graphsignal.inference_span import InferenceSpan
from graphsignal.tracers.operation_profiler import OperationProfiler
from graphsignal.tracers.profiler_utils import create_log_dir, remove_log_dir, find_and_read

logger = logging.getLogger('graphsignal')


class ONNXRuntimeProfiler(OperationProfiler):
    def __init__(self):
        self._is_initialized = None
        self._onnx_version = None
        self._first_trace_path = None
        self._end_lock = threading.Lock()

    def reset(self):
        self._first_trace_path = None

    def start(self, profile, onnx_session):
        if not onnx_session.get_session_options().enable_profiling:
            return

        logger.debug('Activating ONNX profiler')

        # Initialization
        if not self._is_initialized:
            self._is_initialized = True

            self._onnx_version = profiles_pb2.SemVer()
            parse_semver(self._onnx_version, onnxruntime.__version__)

        # Profiler info
        profile.profiler_info.operation_profiler_type = profiles_pb2.ProfilerInfo.ProfilerType.ONNX_PROFILER

        # Framework info
        framework = profile.frameworks.add()
        framework.type = profiles_pb2.FrameworkInfo.FrameworkType.ONNX_FRAMEWORK
        framework.version.CopyFrom(self._onnx_version)

    def stop(self, profile, onnx_session):
        if not onnx_session.get_session_options().enable_profiling:
            return

        logger.debug('Deactivating ONNX profiler')

        # end profiling after first inference and use first trace data 
        # in all other profiles for the current inference session
        with self._end_lock:
            if not self._first_trace_path:
                self._first_trace_path = onnx_session.end_profiling()

        if self._first_trace_path:
            trace_file_size = os.path.getsize(self._first_trace_path)
            if trace_file_size > 50 * 1e6:
                raise Exception('Trace file too big: {0}'.format(trace_file_size))

            with open(self._first_trace_path) as f:
                trace_json = f.read()
                profile.trace_data = gzip.compress(trace_json.encode())


class ONNXRuntimeProfilingSession():
    def __init__(self):
        self.log_dir = '{0}{1}'.format(create_log_dir(), os.path.sep)

    def cleanup(self):
        remove_log_dir(self.log_dir)


_profiler = ONNXRuntimeProfiler()
_profiling_sessions = []


def _cleanup_profiling_sessions():
    if _profiling_sessions:
        for profiling_session in _profiling_sessions:
            profiling_session.cleanup()

atexit.register(_cleanup_profiling_sessions)


def initialize_profiler(onnx_session_options: onnxruntime.SessionOptions):
    graphsignal._check_configured()

    profiling_session = ONNXRuntimeProfilingSession()
    _profiling_sessions.append(profiling_session)

    onnx_session_options.enable_profiling = True
    onnx_session_options.profile_file_prefix = profiling_session.log_dir


def inference_span(
        model_name: str,
        metadata: Optional[dict] = None,
        ensure_profile: Optional[bool] = False,
        onnx_session: Optional[onnxruntime.InferenceSession] = None) -> InferenceSpan:
    graphsignal._check_configured()

    if not onnx_session:
        raise ValueError('onnx_session is required')

    return InferenceSpan(
        model_name=model_name,
        metadata=metadata,
        ensure_profile=ensure_profile,
        operation_profiler=_profiler,
        context=onnx_session)



from typing import Optional
import logging
import os
import sys
import time
import json
import tensorflow as tf

import graphsignal
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import profiles_pb2
from graphsignal.inference_span import InferenceSpan
from graphsignal.tracers.operation_profiler import OperationProfiler
from graphsignal.tracers.profiler_utils import create_log_dir, remove_log_dir, convert_tensorflow_profile

logger = logging.getLogger('graphsignal')

class TensorflowProfiler(OperationProfiler):
    def __init__(self):
        self._is_initialized = False
        self._log_dir = None
        self._tensorflow_version = None
        self._global_rank = None
        self._world_size = None

    def start(self, profile, context):
        logger.debug('Activating TensorFlow profiler')

        # Initialization
        if not self._is_initialized:
            self._is_initialized = True

            self._tensorflow_version = profiles_pb2.SemVer()
            parse_semver(self._tensorflow_version, tf.__version__)
            if compare_semver(self._tensorflow_version, (2, 2, 0)) == -1:
                raise Exception(
                    'TensorFlow profiling is not supported for versions <=2.2')

            if 'TF_CONFIG' in os.environ:
                try:
                    tf_config = json.loads(os.environ['TF_CONFIG'])
                    self._world_size = 0
                    if 'chief' in tf_config['cluster']:
                        self._world_size += len(tf_config['cluster']['chief'])
                    if 'worker' in tf_config['cluster']:
                        self._world_size += len(tf_config['cluster']['worker'])
                    self._global_rank = tf_config['task']['index']
                except:
                    logger.warning('Error parsing TF_CONFIG', exc_info=True)

        # Profiler info
        profile.profiler_info.operation_profiler_type = profiles_pb2.ProfilerInfo.ProfilerType.TENSORFLOW_PROFILER

        # Framework info
        framework = profile.frameworks.add()
        framework.type = profiles_pb2.FrameworkInfo.FrameworkType.TENSORFLOW_FRAMEWORK
        framework.version.CopyFrom(self._tensorflow_version)

        # Process info
        if self._global_rank is not None and self._global_rank >= 0:
            profile.process_usage.global_rank = self._global_rank

        # Cluster stats
        if self._world_size is not None and self._world_size > 0:
            profile.cluster_info.world_size = self._world_size

        try:
            self._log_dir = create_log_dir()

            options = tf.profiler.experimental.ProfilerOptions(
                host_tracer_level=2,
                python_tracer_level=0,
                device_tracer_level=1)
            tf.profiler.experimental.start(self._log_dir, options=options)
        except Exception as e:
            remove_log_dir(self._log_dir)
            raise e

    def stop(self, profile, context):
        logger.debug('Deactivating TensorFlow profiler')

        try:
            tf.profiler.experimental.stop()

            convert_tensorflow_profile(self._log_dir, profile)
        finally:
           remove_log_dir(self._log_dir)


_profiler = TensorflowProfiler()

def inference_span(
        model_name: str,
        metadata: Optional[dict] = None,ensure_profile: Optional[bool] = False) -> InferenceSpan:
    graphsignal._check_configured()

    return InferenceSpan(
        model_name=model_name,
        metadata=metadata,ensure_profile=ensure_profile,
        operation_profiler=_profiler)



from typing import Optional
import logging

from tensorflow import keras
from tensorflow.keras.callbacks import Callback

import graphsignal
from graphsignal.proto import profiles_pb2
from graphsignal.proto_utils import parse_semver
from graphsignal.tracers.tensorflow import TensorflowProfiler
from graphsignal.inference_span import InferenceSpan

logger = logging.getLogger('graphsignal')


class GraphsignalCallback(Callback):
    def __init__(self, model_name: str, metadata: Optional[dict] = None, batch_size: Optional[int] = None):
        super().__init__()
        self._keras_version = None
        self._profiler = TensorflowProfiler()
        self._span = None
        self._model_name = model_name
        self._metadata = metadata
        self._batch_size = batch_size

    def on_test_begin(self, logs=None):
        self._configure_profiler()

    def on_test_end(self, logs=None):
        graphsignal.upload()

    def on_test_batch_begin(self, batch, logs=None):
        self._start_profiler()

    def on_test_batch_end(self, batch, logs=None):
        self._stop_profiler()

    def on_predict_begin(self, logs=None):
        self._configure_profiler()

    def on_predict_end(self, logs=None):
        graphsignal.upload()

    def on_predict_batch_begin(self, batch, logs=None):
        self._start_profiler()

    def on_predict_batch_end(self, batch, logs=None):
        self._stop_profiler()

    def _configure_profiler(self):
        try:
            self._keras_version = profiles_pb2.SemVer()
            parse_semver(self._keras_version, keras.__version__)
        except Exception:
            logger.error('Error configuring Keras profiler', exc_info=True)

    def _start_profiler(self):
        if not self._span:
            self._span = InferenceSpan(
                model_name=self._model_name,
                metadata=self._metadata,
                operation_profiler=self._profiler)
            self._span.set_count('items', self._batch_size)

    def _stop_profiler(self):
        if self._span:
            if self._span._is_scheduled:
                self._update_profile()
            self._span.stop()
            self._span = None

    def _update_profile(self):
        try:
            profile = self._span._profile

            profile.profiler_info.framework_profiler_type = profiles_pb2.ProfilerInfo.ProfilerType.KERAS_PROFILER

            framework = profile.frameworks.add()
            framework.type = profiles_pb2.FrameworkInfo.FrameworkType.KERAS_FRAMEWORK
            framework.version.CopyFrom(self._keras_version)
        except Exception as exc:
            self._span._add_profiler_exception(exc)



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.profile_scheduler import ProfileScheduler

logger = logging.getLogger('graphsignal')


class ProfileSchedulerTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_ensured(self):
        scheduler = ProfileScheduler()
        for _ in range(ProfileScheduler.MAX_ENSURED_SPANS):
            self.assertTrue(scheduler.lock(ensure=True))
            scheduler.unlock()
        scheduler.unlock()

    def test_predefined(self):
        scheduler = ProfileScheduler()
        first_span = next(iter(scheduler._span_filter))
        for idx in range(ProfileScheduler.PREDEFINED_SPANS[-1]):
            if idx + 1 in scheduler._span_filter:
                self.assertTrue(scheduler.lock())
                scheduler.unlock()
            else:
                self.assertFalse(scheduler.lock())
                scheduler.unlock()
        self.assertFalse(scheduler.lock())
        scheduler.unlock()

    def test_interval(self):
        scheduler = ProfileScheduler()
        scheduler._span_filter = {}
        self.assertFalse(scheduler.lock())
        scheduler.unlock()
        self.assertFalse(scheduler._interval_mode)

        # first interval
        scheduler._start_ts = time.time() - ProfileScheduler.MIN_INTERVAL_SEC - 10
        self.assertTrue(scheduler.lock())
        scheduler.unlock()
        self.assertTrue(scheduler._interval_mode)

        # other intervals
        scheduler._start_ts = time.time() - ProfileScheduler.MIN_INTERVAL_SEC - 10
        scheduler._last_profile_ts = time.time() - ProfileScheduler.MIN_INTERVAL_SEC - 10
        self.assertTrue(scheduler.lock())
        scheduler.unlock()

    def test_concurrent(self):
        scheduler = ProfileScheduler()
        self.assertTrue(scheduler.lock(ensure=True))
        self.assertFalse(scheduler.lock(ensure=True))
        self.assertFalse(scheduler.lock(ensure=True))
        scheduler.unlock()
        self.assertTrue(scheduler.lock(ensure=True))
        scheduler.unlock()




from typing import Any, Union, Optional
import time
import sys
import os
import logging
import threading
import uuid
import hashlib
import atexit

from graphsignal.version import __version__
from graphsignal.agent import Agent
from graphsignal.workload import Workload
from graphsignal.uploader import Uploader
from graphsignal.usage.process_reader import ProcessReader
from graphsignal.usage.nvml_reader import NvmlReader
from graphsignal.inference_span import InferenceSpan
from graphsignal import tracers
from graphsignal.proto import profiles_pb2

logger = logging.getLogger('graphsignal')

_agent = None


def _check_configured():
    global _agent
    if not _agent:
        raise ValueError(
            'Profiler not configured, call graphsignal.configure() first')


def _check_and_set_arg(
        name, value, is_str=False, is_int=False, is_bool=False, required=False, max_len=None):
    env_name = 'GRAPHSIGNAL_{0}'.format(name.upper())

    if not value and env_name in os.environ:
        value = os.environ[env_name]
        if value:
            if is_int:
                try:
                    value = int(value)
                except:
                    raise ValueError('configure: invalid format, expected integer: {0}'.format(name))
            elif is_bool:
                value = bool(value)

    if not value:
        if required:
            raise ValueError('configure: missing argument: {0}'.format(name))
    else:
        if is_str:
            if not isinstance(value, str):
                raise ValueError('configure: invalid format, expected string: {0}'.format(name))
            if max_len and len(value) > max_len:
                raise ValueError('configure: invalid format, string too long (>{0}): {1}'.format(name, max_len))
        elif is_int:
            if not isinstance(value, int):
                raise ValueError('configure: invalid format, expected integer: {0}'.format(name))
        elif is_bool:
            if not isinstance(value, bool):
                raise ValueError('configure: invalid format, expected boolean: {0}'.format(name))

    return value


def configure(
        api_key: Optional[str] = None,
        workload_name: Optional[str] = None,
        debug_mode: Optional[bool] = False,
        disable_op_profiler: Optional[bool] = False) -> None:
    global _agent

    if _agent:
        logger.warning('Profiler already configured')
        return

    debug_mode = _check_and_set_arg('debug_mode', debug_mode, is_bool=True)
    if debug_mode:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.WARNING)
    api_key = _check_and_set_arg('api_key', api_key, is_str=True, required=True)
    workload_name = _check_and_set_arg('workload_name', workload_name,  is_str=True, max_len=50)
    disable_op_profiler = _check_and_set_arg('disable_op_profiler', disable_op_profiler, is_bool=True)

    _agent = Agent()
    _agent.worker_id = _uuid_sha1(size=12)
    _agent.api_key = api_key
    if workload_name:
        _agent.workload_name = workload_name
    _agent.disable_op_profiler = disable_op_profiler
    _agent.debug_mode = debug_mode
    _agent.uploader = Uploader()
    _agent.uploader.configure()
    _agent.process_reader = ProcessReader()
    _agent.process_reader.setup()
    _agent.nvml_reader = NvmlReader()
    _agent.nvml_reader.setup()

    _agent.workload = Workload()
    if workload_name:
        _agent.workload.workload_id = _sha1(workload_name, size=12)
    else:
        _agent.workload.workload_id = _sha1('', size=12)

    atexit.register(shutdown)

    logger.debug('Profiler configured')


def workload() -> Workload:
    _check_configured()

    return _agent.workload


def upload(block=False) -> None:
    _check_configured()

    _agent.workload.upload(block=block)


def shutdown() -> None:
    _check_configured()

    global _agent
    atexit.unregister(shutdown)
    _agent.workload.end(block=True)
    _agent.process_reader.shutdown()
    _agent.nvml_reader.shutdown()
    _agent = None

    logger.debug('Profiler shutdown')


def generate_uuid() -> None:
    return _uuid_sha1()    


def _sha1(text, size=-1):
    sha1_hash = hashlib.sha1()
    sha1_hash.update(text.encode('utf-8'))
    return sha1_hash.hexdigest()[0:size]


def _uuid_sha1(size=-1):
    return _sha1(str(uuid.uuid4()), size)


__all__ = [
    '__version__',
    'configure',
    'workload',
    'upload',
    'shutdown',
    'generate_uuid',
    'tracers'
]


import logging
import threading
import os
import random
import time
import traceback

import graphsignal
from graphsignal.profile_scheduler import ProfileScheduler
from graphsignal.proto import profiles_pb2

logger = logging.getLogger('graphsignal')


class InferenceStats:
    MAX_RESERVOIR_SIZE = 100
    MAX_COUNTERS = 10

    def __init__(self):
        self._update_lock = threading.Lock()
        self._start_sec = int(time.time())
        self.time_reservoir_us = []
        self.inference_counter = {}
        self.exception_counter = {}
        self.extra_counters = {}

    def add_time(self, duration_us):
        with self._update_lock:
            if len(self.time_reservoir_us) < InferenceStats.MAX_RESERVOIR_SIZE:
                self.time_reservoir_us.append(duration_us)
            else:
                self.time_reservoir_us[random.randint(0, InferenceStats.MAX_RESERVOIR_SIZE - 1)] = duration_us

    def inc_inference_counter(self, value, timestamp_us):
        with self._update_lock:
            self._inc_counter(self.inference_counter, value, timestamp_us)

    def inc_exception_counter(self, value, timestamp_us):
        with self._update_lock:
            self._inc_counter(self.exception_counter, value, timestamp_us)

    def inc_extra_counter(self, name, value, timestamp_us):
        with self._update_lock:
            if name not in self.extra_counters:
                if len(self.extra_counters) < InferenceStats.MAX_COUNTERS:
                    counter = self.extra_counters[name] = {}
                else:
                    return
            else:
                counter = self.extra_counters[name]

            self._inc_counter(counter, value, timestamp_us)

    def finalize(self, timestamp_us):
        with self._update_lock:
            self._finalize_counter(self.inference_counter, timestamp_us)
            self._finalize_counter(self.exception_counter, timestamp_us)
            for extra_counter in self.extra_counters.values():
                self._finalize_counter(extra_counter, timestamp_us)

    def _inc_counter(self, counter, value, timestamp_us):
        bucket = int(timestamp_us / 1e6)
        if bucket in counter:
            counter[bucket] += value
            counter[self._start_sec] = 0
        else:
            counter[self._start_sec] = 0
            counter[bucket] = value

    def _finalize_counter(self, counter, timestamp_us):
        end_sec = int(timestamp_us / 1e6)
        if end_sec not in counter:
            counter[end_sec] = 0


class Workload:
    MAX_EXCEPTIONS = 10

    def __init__(self):
        self.workload_id = None
        self.metrics = None
        self.profile_schedulers = {}
        self.inference_stats = {}
        self.exception_stats = {}

    def get_profile_scheduler(self, model_name):
        if model_name in self.profile_schedulers:
            return self.profile_schedulers[model_name]
        else:
            profile_scheduler = self.profile_schedulers[model_name] = ProfileScheduler()
            return profile_scheduler

    def reset_inference_stats(self, name):
        self.inference_stats[name] = InferenceStats()

    def get_inference_stats(self, name):
        if name not in self.inference_stats:
            self.reset_inference_stats(name)

        return self.inference_stats[name]

    def add_exception(self, exc_info, timestamp_us):
        if len(self.exception_stats) > Workload.MAX_EXCEPTIONS:
            return

        exc_type = ''
        if exc_info[0] and hasattr(exc_info[0], '__name__'):
            exc_type = str(exc_info[0].__name__)

        exc_message = ''
        if exc_info[1]:
            exc_message = str(exc_info[1])

        key = '{0}:{1}'.format(exc_type, exc_message)

        if key in self.exception_stats:
            exc_stat = self.exception_stats[key]
            exc_stat.count += 1
        else:
            exc_stat = self.exception_stats[key] = profiles_pb2.ExceptionStats()
            exc_stat.exc_type = exc_type
            exc_stat.message = exc_message
            if exc_info[2]:
                frames = traceback.format_tb(exc_info[2])
                if len(frames) > 0:
                    exc_stat.stack_trace = ''.join(frames)
            exc_stat.count += 1
        exc_stat.update_ms = int(timestamp_us / 1e3)

    def create_profile(self):
        profile = profiles_pb2.MLProfile()
        if graphsignal._agent.workload_name:
            profile.workload_name = graphsignal._agent.workload_name
        profile.worker_id = graphsignal._agent.worker_id
        profile.workload_id = self.workload_id
        return profile

    def upload(self, block=False):
        if block:
            graphsignal._agent.uploader.flush()
        else:
            graphsignal._agent.uploader.flush_in_thread()

    def tick(self, block=False):
        self.upload(block=False)

    def end(self, block=False):
        self.upload(block=block)



from typing import Optional
import logging
import os
import sys
import time
import gzip
import jax

import graphsignal
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import profiles_pb2
from graphsignal.inference_span import InferenceSpan
from graphsignal.tracers.operation_profiler import OperationProfiler
from graphsignal.tracers.profiler_utils import create_log_dir, remove_log_dir, convert_tensorflow_profile

logger = logging.getLogger('graphsignal')

class JaxProfiler(OperationProfiler):
    def __init__(self):
        self._is_initialized = False
        self._log_dir = None
        self._jax_version = None

    def start(self, profile, context):
        logger.debug('Activating JAX profiler')

        # Initialization
        if not self._is_initialized:
            self._is_initialized = True

            self._jax_version = profiles_pb2.SemVer()
            parse_semver(self._jax_version, jax.__version__)

        # Profiler info
        profile.profiler_info.operation_profiler_type = profiles_pb2.ProfilerInfo.ProfilerType.JAX_PROFILER

        # Framework info
        framework = profile.frameworks.add()
        framework.type = profiles_pb2.FrameworkInfo.FrameworkType.JAX_FRAMEWORK
        framework.version.CopyFrom(self._jax_version)

        try:
            self._log_dir = create_log_dir()

            jax.profiler.start_trace(self._log_dir)
        except Exception as e:
            remove_log_dir(self._log_dir)
            raise e

    def stop(self, profile, context):
        logger.debug('Deactivating JAX profiler')

        try:
            jax.profiler.stop_trace()

            convert_tensorflow_profile(self._log_dir, profile)
        finally:
            remove_log_dir(self._log_dir)


_profiler = JaxProfiler()

def inference_span(
        model_name: str,
        metadata: Optional[dict] = None,ensure_profile: Optional[bool] = False) -> InferenceSpan:
    graphsignal._check_configured()

    return InferenceSpan(
        model_name=model_name,
        metadata=metadata,ensure_profile=ensure_profile,
        operation_profiler=_profiler)
